# xDevs.com Calibration app for Keithley 2304, v0.1
# https://xdevs.com/fix/kei2304a
# https://xdevs.com/fix/hp3458a
import sys
import time
import Gpib
import signal
import numbers
from math import log10

dmm_val = 1.0
dmm_temp = 36.6

class Timeout():
  """Timeout class using ALARM signal"""
  class Timeout(Exception): pass

  def __init__(self, sec):
    self.sec = sec

  def __enter__(self):
    signal.signal(signal.SIGALRM, self.raise_timeout)
    signal.alarm(self.sec)

  def __exit__(self, *args):
    signal.alarm(0) # disable alarm

  def raise_timeout(self, *args):
    raise Timeout.Timeout()

class dmm():
    global dmm_val 
    data = ""
    temp = 36.6
    cnti = 12
    
    def __init__(self,gpib,refhp,name):
        self.gpib = gpib
        self.inst = Gpib.Gpib(0,self.gpib, timeout=60) # 3458A GPIB Address = self.gpib
        self.refhp = refhp
        self.name = name
        self.init_inst()

    def init_inst(self):
        # Setup HP 3458A
        self.inst.write ("PRESET NORM")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCV,AUTO")
        self.inst.write ("TARM HOLD")
        self.inst.write ("TRIG LINE")
        self.inst.write ("NPLC 100")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("MEM OFF")
        self.inst.write ("END ALWAYS")
        self.inst.write ("NDIG 9")

    def switch_dci(self):
        self.inst.write ("PRESET NORM")
        self.inst.write ("NPLC 100")
        self.inst.write ("AZERO ON")
        self.inst.write ("NRDGS 1,AUTO")
        self.inst.write ("OFORMAT ASCII")
        self.inst.write ("FUNC DCI,AUTO")
        self.inst.write ("NDIG 9")
        
    def read_data(self,cmd):
        data_float = 0.0
        data_str = ""
        self.inst.write(cmd)
        
        try:
            with Timeout(20):
                data_str = self.inst.read()
        except Timeout.Timeout:
            print ("Timeout exception from dmm %s on read_data() inst.read()\n" % self.name)
            return (0,float(0))
        #print ("Reading from dmm %s = %s" % (self.name,data_str))
        try:
            data_float = float(data_str)
        except ValueError:
            print("Exception thrown by dmm %s on read_data() - ValueError = %s\n" % (self.name,data_str))
            return (0,float(0)) # Exception on float conversion, 0 = error
        return (1,data_float) # Good read, 1 = converted to float w/o exception
        
    def get_temp(self):
        global dmm_temp
        print("Reading DMM temp")
        self.inst.write("NPLC 10")                          # Switch to NPLC10 for quick readout
        self.inst.write("TARM SGL,1")
        self.temp_status_flag,temp = self.read_data("TEMP?")
        if (self.temp_status_flag):
            self.temp = temp
        dmm_temp = temp
        self.inst.write ("NPLC 100")                        # Switch back NPLC100
        return self.temp
        
    def get_temp_status(self):
        return self.temp_status_flag

    def get_data(self):
        global dmm_val
        self.status_flag,data = self.read_data("TARM SGL,1")
        if (self.status_flag):
            self.data = data
            dmm_val = float(data)
            self.ppm = ((float(self.data) / self.refhp)-1)*1E6
        return self.data
        
    def exec_acal(self):
        sys.stdout.write ("\033[1;35mACAL ALL procedure start, please wait 12 minutes ")
        self.inst.write("ACAL ALL")
        for cnt in range(0 ,48):        # wait 720 seconds, print dot every 15s
            sys.stdout.write ("\033[0;43m*")
            time.sleep(15)
        print("\r\nACAL procedure done...\033[1;39m")
    
    def exec_idn(self):
        self.inst.write ("END ALWAYS")
        self.inst.write ("ID?")
        dat = self.inst.read()
        tstr = dat.split()
        if (tstr[0] == "HP3458A"):
            sys.stdout.write ("\r\n\033[1;32m%s detected...\033[1;39m" % tstr[0])
        else:
            sys.stdout.write ("\r\n\033[1;31mNo DMM present, exiting!\033[1;39m")
            quit()

    def get_data_status(self):
        return self.status_flag

with open('callog_psu.txt','wb') as b:
    kei2304_idn = "KEITHLEY INSTRUMENTS INC."
    kei2304_mdl = "MODEL 2304A"
    orig_val = []
    tval = 0.0
    res4 = 3.99895
    res4k = 4000.000
    kei_error = 0
    plc = 60 #60 Hz
    cnt = 0
    samples = 0
    nplc = 0
    val = 3.423
    meas = 0.0
    temp = 36.6
    
    kei2304 = Gpib.Gpib(0,5)               # Keithley Model 2400 GPIB address init 5
    kei2304.clear()

    for cnt in range(30):
        orig_val.append(0)
        
    sys.stdout.write("\033[0;36m Keithley 2304A PSU calibration tool \r\n  Using NI GPIB adapter with next instruments config: \r\n * GPIB 5 : Keithley 2304A PSU \r\n * GPIB 3 : HP 3458A 8.5-digit DMM \r\n ! Do not swap terminals during calibration unless asked!\033[0;39m\r\n")
    
    kei2304.write ("*IDN?")
    idn_str = kei2304.read()
    idmfg = idn_str.split(",")
    if (kei2304_idn == idmfg[0]):
        if (kei2304_mdl == idmfg[1]):
            sys.stdout.write("\033[1;31mKeithley \033[0;32mModel 2304A - detected, S/N %s, Version: %s\033[0;39m\r\n" % (idmfg[2], idmfg[3]))
            b.write("\r\nKeithley Model 2304A - detected, S/N %s, Version: %s\r\n" % (idmfg[2], idmfg[3]))
        else:
            print ("\033[0;31mIncorrect Model! Check GPIB address. Calibration cancelled.\033[0;39m")    
            quit()
    else:
        sys.stdout.write ("\033[0;31mNo Keithley instrument detected. Check GPIB address. Calibration cancelled.\033[0;39m")
        quit()
    
    # Keithley Model 2304A detected OK, proceed with perf-val test
    
    kei2304.write ("*CLS")
    kei2304.write ("*ESE 1;*SRE 32")
    kei2304.write ("*RST")
    kei2304.write (":SENS:FUNC:CONC OFF")
    kei2304.write (":SENS:FUNC 'VOLT:DC'")
    kei2304.write (":SOUR:FUNC VOLT")
    kei2304.write (":OUTP:STAT OFF")

    print "Performance verification for Model 2304 - Voltage source"
    b.write ("Pre-cal K2304 performance verification data - voltage source \r\n")

    dmm = dmm(3,dmm_val,"3458A") # GPIB 3
    dmm.exec_idn()
    dmm.get_temp()
    print ("\033[0;33mHP3458A TEMP = %2.1f C\033[0;33m" % dmm_temp)
    b.write ("HP 3458A detected\r\n TEMP? = %2.1f C\r\n" % dmm_temp)
        
    while True:
        res4 = float(raw_input("\033[1;33mEnter value of reference 4R resistor, in ohms, e.g. 3.99895 and press Enter...\033[1;39m"))
        if ((res4 >= 3.600) & (res4 < 4.400)):
            print ("Resistance entered = %f Ohm" % res4)
            b.write ("Resistance 4R entered = %f Ohm\r\n" % res4)
            break
        print ("Incorrect resistance, please use 3.6 > R > 4.4!")
    
    while True:
        res4k = float(raw_input("\033[1;33mEnter value of 4000R resistor, in ohms, e.g. 3992.3 and press Enter...\033[1;39m"))
        if ((res4k >= 3800) & (res4k < 4200)):
            print ("Resistance entered = %f Ohm" % res4k)
            b.write ("Resistance 4000R entered = %f Ohm\r\n" % res4k)
            break
        print ("Incorrect resistance, please use 3800 > R > 4200!")
    
    print ("\033[1;31m Connect DMM volts input to SOURCE and DVM IN terminals. \033[1;39m")
    raw_input("\033[1;33mPress Enter to continue with pre-cal test...\033[1;39m")
    
    dmm.exec_acal()         # Execute ACAL on HP 3458A for best accuracy. Takes 12 minutes to complete.
    
    #Ramp 0 to 20V in 0.1 steps.
    #Measure each point
    #Report deviation
    
    kei2304.write (":OUTP:STAT ON")                         # PSU output enabled
    for cnt in range (0, 200):                              # 200 points to sample
        
        tval = float(cnt) / 10.0
        print ("Setting PSU output %5.3f" % tval)
        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
        kei2304.write(":MEAS:VOLT?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:VOLT?")
        psuval = float(kei2304.read())
    
        ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")                        # PSU output disabled

    # DVM input accuracy test
    
    kei2304.write (":OUTP:STAT OFF")
    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

    print "Performance verification for Model 2304 DVM IN"
    b.write("Pre-cal K2304 performance verification data - DVM IN \r\n")

    kei2304.write (":OUTP:STAT ON")
    kei2304.write(":SENS:FUNC DVM")
    for cnt in range (0, 200):
        
        tval = float(cnt) / 10.0
        print ("Setting PSU output %5.3f" % tval)
        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
        kei2304.write(":MEAS:DVM?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:DVM?")
        psuval = float(kei2304.read())
    
        ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")

    # PSU output current compliance test
    
    print "Performance verification for Model 2304 - Current compliance 5A"
    b.write("Pre-cal K2304 performance verification data - current compliance 5A \r\n")
    print ("Connect 4OHM resistor to PSU source and sense HI/LO and DMM voltage HI/LO")
    
    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

    kei2304.write (":OUTP:STAT ON")
    for cnt in range (1, 6):
        tval = float(cnt) + 0.00000001
    kei2304.write (":OUTP:STAT ON")
    print ("Setting PSU output %5.3f" % tval)
    kei2304.write(":SENS:CURR:RANG MAX")
    kei2304.write(":SOUR:VOLT 20")
    kei2304.write(":SOUR:CURR:LIM:VAL %5.3f" % tval)

        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCI %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        val = val / res4                                    # convert to current
        ppmd = ((val / float(tval) + 0.0000001) - 1) * 1E6
        
        kei2304.write(":MEAS:CURR?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:CURR?")
        psuval = float(kei2304.read())

        ppmv = ((psuval / val + 0.000001) - 1) * 1E6

        kei2304.write (":OUTP:STAT OFF")
        print ("Sleeping 10 seconds to allow resistors to cool")
        time.sleep(10)                            
       
        print ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")
    
    # PSU calibration procedure

    raw_input("\033[1;33mPress Enter to start cal or CTRL-C to abort...\033[1;39m")
    print "Calibration for Model 2304"
    b.write("Calibration Model 2304 started\r\n")

    kei2304.write (":CAL:PROT:CODE 'KI002304'")
    kei2304.write (":CAL:PROT:INIT")
    print ("Step 0, 19V output")
    b.write ("Step 0, 19V output\r\n")
    kei2304.write (":CAL:PROT:STEP0 19")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 10):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 10
    ppmd = (((val / 19) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Deviation source %f ppm" % (val, ppmd))
    b.write ("Measured Vout = %f VDC, Deviation source %f ppm\r\n" % (val, ppmd))
    kei2304.write (":CAL:PROT:STEP1 %8.6f" % val)

    print ("Step 2, 19V DVM output")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 10):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 10
    ppmd = (((val / 19) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Deviation source %f ppm" % (val, ppmd))
    b.write ("Measured Vout = %f VDC, Deviation source %f ppm\r\n" % (val, ppmd))
    kei2304.write (":CAL:PROT:STEP2 %8.6f" % val)

    kei2304.write (":CAL:PROT:STEP3")
    print("STEP 3, connect 4 ohm")
    raw_input("\033[1;33mPress Enter to continue with cal...\033[1;39m")

    kei2304.write (":CAL:PROT:STEP4 1.9")

    print ("Step 4, 1.9A over 4R")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4
    ppmd = (((val / 1.9) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP5 %8.6f" % ccurr)

    print ("Step 5, 1.9A over 4R")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4
    ppmd = (((val / 1.9) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP6 %8.6f" % ccurr)

    print("STEP 6, connect 4000 ohm")
    raw_input("\033[1;33mPress Enter to continue with cal...\033[1;39m")
    print ("Step 7, 5mA over 4K resistor")
    kei2304.write (":CAL:PROT:STEP7")

    val = 0 
    time.sleep(0.1)
    for samples in range (0, 5):
        dmm.get_data()
        print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
        val += dmm_val
    val = val / 5
    ccurr = val / res4k
    ppmd = (((val / 0.005) - 1) * 1E6)

    print ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm" % (val, ccurr, ppmd))
    b.write ("Measured Vout = %f VDC, Current = %f A, Deviation source %f ppm\r\n" % (val, ccurr, ppmd))
    kei2304.write (":CAL:PROT:STEP8 %8.6E" % ccurr)

    raw_input("\033[1;33mPress Enter to save cal...\033[1;39m")
    print ("Saving calibration with date 2016 04 11")
    b.write ("Saving calibration with date 2016 04 11\r\n")
    
    kei2304.write (":CAL:PROT:DATE 2016,04,11")
    kei2304.write (":CAL:PROT:SAVE")
    kei2304.write (":CAL:PROT:LOCK")
    kei2304.write (":OUTP:STAT OFF")

    print ("Calibration complete")
    b.write ("Calibration complete \r\n")
    
    # Post-cal performance test
    
    #Ramp 0 to 20V in 0.1 steps.
    #Measure each point
    #Report deviation
    
    print ("Post-calibration performance test \r\n\033[1;31m Connect DMM volts input to SOURCE and DVM IN terminals. \033[1;39m")
    raw_input("\033[1;33mPress Enter to continue with post-cal test...\033[1;39m")
        
    kei2304.write (":OUTP:STAT ON")                         # PSU output enabled
    for cnt in range (0, 200):                              # 200 points to sample
        
        tval = float(cnt) / 10.0
        print ("Setting PSU output %5.3f" % tval)
        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
        kei2304.write(":MEAS:VOLT?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:VOLT?")
        psuval = float(kei2304.read())
    
        ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")                        # PSU output disabled

    # DVM input accuracy test
    
    kei2304.write (":OUTP:STAT OFF")
    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

    print "Performance verification for Model 2304 DVM IN"
    b.write("Pre-cal K2304 performance verification data - DVM IN \r\n")

    kei2304.write (":OUTP:STAT ON")
    kei2304.write(":SENS:FUNC DVM")
    for cnt in range (0, 200):
        
        tval = float(cnt) / 10.0
        print ("Setting PSU output %5.3f" % tval)
        kei2304.write(":SOUR:VOLT %5.3f" % tval)
        
        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCV %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        ppmd = ((val / (tval + 0.0000000001)) - 1) * 1E6
    
        kei2304.write(":MEAS:DVM?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:DVM?")
        psuval = float(kei2304.read())
    
        ppmv = ((psuval / (val + 0.0000000001)) - 1) * 1E6

        print ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Vout = %f VDC, Deviation source %f ppm, PSU fetch: %f VDC, Delta-DMM %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")

    # PSU output current compliance test
    
    print "Performance verification for Model 2304 - Current compliance 5A"
    b.write("Pre-cal K2304 performance verification data - current compliance 5A \r\n")
    print ("Connect 4OHM resistor to PSU source and sense HI/LO and DMM voltage HI/LO")
    
    raw_input("\033[1;33mPress Enter to continue with test...\033[1;39m")

    kei2304.write (":OUTP:STAT ON")
    for cnt in range (1, 6):
        tval = float(cnt) + 0.00000001
    kei2304.write (":OUTP:STAT ON")
    print ("Setting PSU output %5.3f" % tval)
    kei2304.write(":SENS:CURR:RANG MAX")
    kei2304.write(":SOUR:VOLT 20")
    kei2304.write(":SOUR:CURR:LIM:VAL %5.3f" % tval)

        val = 0 
        time.sleep(0.1)
        for samples in range (0, 3):
            dmm.get_data()
            print ("Measured DCI %d : %.9E VDC" % (samples, dmm_val))
            val += dmm_val
        val = val / 3
        val = val / res4                                    # convert to current
        ppmd = ((val / float(tval) + 0.0000001) - 1) * 1E6
        
        kei2304.write(":MEAS:CURR?")
        psuval = float(kei2304.read())
        kei2304.write(":MEAS:CURR?")
        psuval = float(kei2304.read())

        ppmv = ((psuval / val + 0.000001) - 1) * 1E6

        kei2304.write (":OUTP:STAT OFF")
        print ("Sleeping 10 seconds to allow resistors to cool")
        time.sleep(10)                            
       
        print ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm" % (val, ppmd, psuval, ppmv))
        b.write ("Measured Iout = %f DCI, Deviation source %f ppm, PSU fetch: %f ADC, Delta %f ppm\r\n" % (val, ppmd, psuval, ppmv))
    kei2304.write (":OUTP:STAT OFF")
    
    b.close()
    quit()
